@@ -7,24 +7,50 @@
 
 #include "ibusfrontend.h"
 #include <fcntl.h>
-#include <signal.h>
+#include <sys/stat.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <fstream>
-#include <mutex>
+#include <cerrno>
+#include <csignal>
+#include <cstddef>
+#include <cstdint>
+#include <cstdio>
+#include <cstdlib>
+#include <memory>
+#include <optional>
+#include <set>
+#include <string>
+#include <tuple>
+#include <utility>
+#include <vector>
+#include <fmt/core.h>
 #include <fmt/format.h>
-#include <uuid/uuid.h>
 #include "fcitx-config/iniparser.h"
+#include "fcitx-config/rawconfig.h"
+#include "fcitx-utils/capabilityflags.h"
+#include "fcitx-utils/dbus/bus.h"
 #include "fcitx-utils/dbus/message.h"
 #include "fcitx-utils/dbus/objectvtable.h"
 #include "fcitx-utils/dbus/servicewatcher.h"
 #include "fcitx-utils/dbus/variant.h"
+#include "fcitx-utils/event.h"
+#include "fcitx-utils/flags.h"
+#include "fcitx-utils/handlertable.h"
+#include "fcitx-utils/key.h"
 #include "fcitx-utils/log.h"
-#include "fcitx-utils/metastring.h"
+#include "fcitx-utils/macros.h"
+#include "fcitx-utils/misc.h"
+#include "fcitx-utils/rect.h"
 #include "fcitx-utils/standardpath.h"
 #include "fcitx-utils/stringutils.h"
+#include "fcitx-utils/textformatflags.h"
 #include "fcitx-utils/utf8.h"
+#include "fcitx-utils/uuid_p.h"
+#include "fcitx/addonfactory.h"
+#include "fcitx/addoninstance.h"
+#include "fcitx/event.h"
 #include "fcitx/inputcontext.h"
+#include "fcitx/inputpanel.h"
 #include "fcitx/instance.h"
 #include "fcitx/misc_p.h"
 #include "dbus_public.h"
@@ -36,17 +62,15 @@ FCITX_DEFINE_LOG_CATEGORY(ibus, "ibus")
 #define IBUS_INPUTMETHOD_DBUS_INTERFACE "org.freedesktop.IBus"
 #define IBUS_INPUTCONTEXT_DBUS_INTERFACE "org.freedesktop.IBus.InputContext"
 #define IBUS_SERVICE_DBUS_INTERFACE "org.freedesktop.IBus.Service"
-#define IBUS_PANEL_SERVICE_NAME "org.freedesktop.IBus.Panel"
+#define IBUS_PANEL_SERVICE "org.freedesktop.IBus.Panel"
+#define IBUS_SERVICE "org.freedesktop.IBus"
+#define IBUS_PORTAL_DBUS_SERVICE "org.freedesktop.portal.IBus"
+#define IBUS_PORTAL_DBUS_INTERFACE "org.freedesktop.IBus.Portal"
 
 namespace fcitx {
 
 namespace {
 
-bool isInFlatpak() {
-    static bool inFlatpak = fs::isreg("/.flatpak-info");
-    return inFlatpak;
-}
-
 std::string getSocketPath(bool isWayland) {
     auto *path = getenv("IBUS_ADDRESS_FILE");
     if (path) {
@@ -90,9 +114,10 @@ std::string getSocketPath(bool isWayland) {
                             hostname, "-", displaynumber));
 }
 
-std::string getFullSocketPath(bool isWayland) {
+std::string getFullSocketPath(const StandardPath &standardPath,
+                              bool isWayland) {
     return stringutils::joinPath(
-        StandardPath::global().userDirectory(StandardPath::Type::Config),
+        standardPath.userDirectory(StandardPath::Type::Config),
         getSocketPath(isWayland));
 }
 
@@ -131,8 +156,8 @@ std::pair<std::string, pid_t> getAddress(const std::string &socketPath) {
 }
 
 pid_t runIBusExit() {
-    pid_t child_pid;
-    if ((child_pid = fork()) == -1) {
+    pid_t child_pid = fork();
+    if (child_pid == -1) {
         perror("fork");
         return -1;
     }
@@ -242,11 +267,6 @@ public:
                            }
                        })),
           name_(sender) {
-        processKeyEventMethod.setClosureFunction(
-            [this](dbus::Message message, const dbus::ObjectMethod &method) {
-                InputContextEventBlocker blocker(this);
-                return method(std::move(message));
-            });
         im->bus()->addObjectVTable(path().path(),
                                    IBUS_INPUTCONTEXT_DBUS_INTERFACE, *this);
         im->bus()->addObjectVTable(path().path(), IBUS_SERVICE_DBUS_INTERFACE,
@@ -254,7 +274,7 @@ public:
         created();
     }
 
-    ~IBusInputContext() { InputContext::destroy(); }
+    ~IBusInputContext() override { InputContext::destroy(); }
 
     const char *frontend() const override { return "ibus"; }
 
@@ -343,7 +363,7 @@ public:
 
     void resetDBus() {
         CHECK_SENDER_OR_RETURN;
-        reset(ResetReason::Client);
+        reset();
     }
 
     void setCursorLocation(int x, int y, int w, int h) {
@@ -413,8 +433,11 @@ public:
     void enable() {}
     void disable() {}
     static bool isEnabled() { return true; }
-    void propertyActivate(const std::string &, int32_t) {}
-    void setEngine(const std::string &) {}
+    void propertyActivate(const std::string &name, int32_t state) {
+        FCITX_UNUSED(name);
+        FCITX_UNUSED(state);
+    }
+    void setEngine(const std::string &engine) { FCITX_UNUSED(engine); }
     static dbus::Variant getEngine() { return dbus::Variant(0); }
     void setSurroundingText(const dbus::Variant &text, uint32_t cursor,
                             uint32_t anchor) {
@@ -515,7 +538,7 @@ private:
 
 #define CASE_PURPOSE(_PURPOSE, _CAPABILITY)                                    \
     case _PURPOSE:                                                             \
-        flag |= _CAPABILITY;                                                   \
+        flag |= (_CAPABILITY);                                                 \
         break;
 
         switch (purpose) {
@@ -551,8 +574,8 @@ private:
         };
 
 #define CHECK_HINTS(_HINTS, _CAPABILITY)                                       \
-    if (hints & _HINTS) {                                                      \
-        flag |= _CAPABILITY;                                                   \
+    if (hints & (_HINTS)) {                                                    \
+        flag |= (_CAPABILITY);                                                 \
     }
 
         CHECK_HINTS(GTK_INPUT_HINT_SPELLCHECK,
@@ -588,12 +611,29 @@ private:
             clientCommitPreedit_ = std::get<0>(value);
         }),
         dbus::PropertyOption::Hidden);
+    FCITX_OBJECT_VTABLE_WRITABLE_PROPERTY(
+        effectivePostProcessKeyEvent, "EffectivePostProcessKeyEvent", "(b)",
+        ([this]() -> dbus::DBusStruct<bool> {
+            return {effectivePostProcessKeyEvent_};
+        }),
+        ([this](dbus::DBusStruct<bool> value) {
+            effectivePostProcessKeyEvent_ = std::get<0>(value);
+        }),
+        dbus::PropertyOption::Hidden);
+    FCITX_OBJECT_VTABLE_PROPERTY(
+        postProcessKeyEvent, "PostProcessKeyEvent", "(a(yv))",
+        ([]() -> dbus::DBusStruct<
+                  std::vector<dbus::DBusStruct<uint8_t, dbus::Variant>>> {
+            return {};
+        }),
+        dbus::PropertyOption::Hidden);
 
     dbus::ObjectPath path_;
     IBusFrontend *im_;
     std::unique_ptr<HandlerTableEntry<dbus::ServiceWatcherCallback>> handler_;
     std::string name_;
     bool clientCommitPreedit_ = false;
+    bool effectivePostProcessKeyEvent_ = false;
 
     IBusService service_{this};
 };
@@ -614,16 +654,13 @@ IBusFrontend::createInputContext(const std::string & /* unused */) {
     return ic->path();
 }
 
-#define IBUS_PORTAL_DBUS_SERVICE "org.freedesktop.portal.IBus"
-#define IBUS_PORTAL_DBUS_INTERFACE "org.freedesktop.IBus.Portal"
-
-std::set<std::string> allSocketPaths() {
+std::set<std::string> allSocketPaths(const StandardPath &standardPath) {
     std::set<std::string> paths;
     if (isInFlatpak()) {
         // Flatpak always use DISPLAY=:99, which means we will need to guess
         // what files are available.
-        auto map = StandardPath::global().multiOpenFilter(
-            StandardPath::Type::Config, "ibus/bus", O_RDONLY,
+        auto map = standardPath.locateWithFilter(
+            StandardPath::Type::Config, "ibus/bus",
             [](const std::string &path, const std::string &, bool user) {
                 if (!user) {
                     return false;
@@ -632,14 +669,14 @@ std::set<std::string> allSocketPaths() {
             });
 
         for (const auto &item : map) {
-            paths.insert(item.second.path());
+            paths.insert(item.second);
         }
 
         // Make the guess that display is 0, it is the most common value that
         // people would have.
         if (paths.empty()) {
-            auto configHome = StandardPath::global().userDirectory(
-                StandardPath::Type::Config);
+            auto configHome =
+                standardPath.userDirectory(StandardPath::Type::Config);
             if (!configHome.empty()) {
                 paths.insert(stringutils::joinPath(
                     configHome, "ibus/bus",
@@ -647,24 +684,62 @@ std::set<std::string> allSocketPaths() {
             }
         }
     } else {
-        if (auto path = getFullSocketPath(false); !path.empty()) {
+        if (auto path = getFullSocketPath(standardPath, false); !path.empty()) {
             paths.insert(std::move(path));
         }
     }
 
     // Also add wayland.
-    if (auto path = getFullSocketPath(true); !path.empty()) {
+    if (auto path = getFullSocketPath(standardPath, true); !path.empty()) {
         paths.insert(std::move(path));
     }
     return paths;
 }
 
 IBusFrontendModule::IBusFrontendModule(Instance *instance)
-    : instance_(instance), socketPaths_(allSocketPaths()) {
+    : instance_(instance), socketPaths_(allSocketPaths(standardPath_)) {
     dbus::VariantTypeRegistry::defaultRegistry().registerType<IBusText>();
     dbus::VariantTypeRegistry::defaultRegistry().registerType<IBusAttribute>();
     dbus::VariantTypeRegistry::defaultRegistry().registerType<IBusAttrList>();
-    replaceIBus();
+
+    // Do the resource initialization first.
+    inputMethod1_ = std::make_unique<IBusFrontend>(
+        this, bus(), IBUS_INPUTMETHOD_DBUS_INTERFACE);
+    portalBus_ = std::make_unique<dbus::Bus>(bus()->address());
+    portalIBusFrontend_ = std::make_unique<IBusFrontend>(
+        this, portalBus_.get(), IBUS_PORTAL_DBUS_INTERFACE);
+    portalBus_->attachEventLoop(&instance_->eventLoop());
+
+    FCITX_IBUS_DEBUG() << "Requesting IBus service name.";
+    if (!bus()->requestName(
+            IBUS_SERVICE,
+            Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
+                                         dbus::RequestNameFlag::Queue})) {
+        FCITX_IBUS_WARN() << "Failed to request IBus service name.";
+        return;
+    }
+
+    bus()->requestName(
+        IBUS_PANEL_SERVICE,
+        Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
+                                     dbus::RequestNameFlag::Queue});
+
+    if (!portalBus_->requestName(
+            IBUS_PORTAL_DBUS_SERVICE,
+            Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
+                                         dbus::RequestNameFlag::Queue})) {
+        FCITX_IBUS_WARN() << "Can not get portal ibus name right now.";
+    }
+
+    // Add some delay for the initial check, since xcb may trigger a ibus
+    // restart on GNOME.
+    timeEvent_ = instance->eventLoop().addTimeEvent(
+        CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1000000, 0,
+        [this](EventSourceTime *, uint64_t) {
+            replaceIBus(/*recheck=*/true);
+            return true;
+        });
+    ;
 }
 
 IBusFrontendModule::~IBusFrontendModule() {
@@ -672,17 +747,17 @@ IBusFrontendModule::~IBusFrontendModule() {
         portalBus_->releaseName(IBUS_PORTAL_DBUS_SERVICE);
     }
 
-    if (addressWrote_.empty()) {
+    if (addressWrote_.empty() || socketPaths_.empty()) {
         return;
     }
+    // Writeback an empty invalid address file.
+    RawConfig config;
+    config.setValueByPath("IBUS_ADDRESS", "");
+    config.setValueByPath("IBUS_DAEMON_PID", "");
     for (const auto &path : socketPaths_) {
         auto address = getAddress(path);
         if (address.first == addressWrote_ && address.second == pidWrote_) {
-            // Writeback an empty invalid address file.
-            RawConfig config;
-            config.setValueByPath("IBUS_ADDRESS", "");
-            config.setValueByPath("IBUS_DAEMON_PID", "");
-            StandardPath::global().safeSave(
+            standardPath_.safeSave(
                 StandardPath::Type::Config, path,
                 [&config](int fd) { return writeAsIni(config, fd); });
         }
@@ -693,123 +768,160 @@ dbus::Bus *IBusFrontendModule::bus() {
     return dbus()->call<IDBusModule::bus>();
 }
 
-void IBusFrontendModule::replaceIBus() {
-    FCITX_IBUS_DEBUG() << "Found ibus socket files: " << socketPaths_;
-    std::pair<std::string, pid_t> address;
-    for (const auto &path : socketPaths_) {
-        address = getAddress(path);
+std::optional<std::pair<std::string, pid_t>>
+readIBusInfo(const std::string &socketPath) {
+    std::pair<std::string, pid_t> address = getAddress(socketPath);
 
-        FCITX_IBUS_DEBUG() << "Found ibus address from file " << path << ": "
-                           << address;
+    FCITX_IBUS_DEBUG() << "Found ibus address from file " << socketPath << ": "
+                       << address;
 
-        if (isInFlatpak()) {
-            // Check the in flatpak special pid value.
-            if (address.second == 0) {
-                continue;
-            }
-        } else {
-            // It's not meaningful to compare pid from different pid namespace.
-            if (address.second == getpid()) {
-                continue;
-            }
+    if (isInFlatpak()) {
+        // Check the in flatpak special pid value.
+        if (address.second == 0) {
+            return std::nullopt;
         }
-        if (!address.first.empty() &&
-            address.first.find("fcitx_random_string") == std::string::npos) {
-            break;
+    } else {
+        // It's not meaningful to compare pid from different pid namespace.
+        if (address.second == getpid()) {
+            return std::nullopt;
         }
     }
-    const auto &oldAddress = address.first;
-    FCITX_IBUS_DEBUG() << "Old ibus address is: " << oldAddress;
-    if (!oldAddress.empty()) {
-        if (isInFlatpak()) {
-            // When running inside flatpak, ibus command won't be available.
-            // sd-bus does not connect to IBus's dbus, probably due to different
-            // implemenation on dbus protocol. sd-bus has bug on xdg-dbus-proxy
-            // anyway, so luckily, in the flatpak we are forced to used libdbus
-            // which works for bus of ibus.
-            FCITX_IBUS_DEBUG() << "Connecting to ibus address: " << oldAddress;
-            dbus::Bus bus(oldAddress);
-            if (bus.isOpen()) {
-                auto call = bus.createMethodCall(
-                    "org.freedesktop.IBus", "/org/freedesktop/IBus",
-                    "org.freedesktop.IBus", "Exit");
-                call << false;
-                call.call(1000000);
-                // Wait 1 second to become ibus.
-                timeEvent_ = instance()->eventLoop().addTimeEvent(
-                    CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1000000, 0,
-                    [this](EventSourceTime *, uint64_t) {
-                        becomeIBus();
-                        return true;
-                    });
-            }
-        } else {
-            // If ibus command is not available, then ibus is probably not
-            // installed anyway, so there is nothing to worry about.
-            auto pid = runIBusExit();
-            if (pid > 0) {
-                FCITX_IBUS_DEBUG() << "Running ibus exit.";
-                timeEvent_ = instance()->eventLoop().addTimeEvent(
-                    CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1000000, 0,
-                    [this, pid, address](EventSourceTime *, uint64_t) {
-                        int stat = -1;
-                        pid_t ret;
-                        while ((ret = waitpid(pid, &stat, WNOHANG)) <= 0) {
-                            if (ret == 0) {
-                                FCITX_IBUS_DEBUG()
-                                    << "ibus exit haven't ended yet, kill it.";
-                                kill(pid, SIGKILL);
-                                waitpid(pid, &stat, WNOHANG);
-                                break;
-                            }
-
-                            if (errno != EINTR) {
-                                stat = -1;
-                                break;
-                            }
-                        }
-
-                        FCITX_IBUS_DEBUG() << "ibus exit returns with " << stat;
-                        if (stat != 0) {
-                            auto cmd = readFileContent(stringutils::joinPath(
-                                "/proc", address.second, "cmdline"));
-                            if (cmd.find("ibus-daemon") != std::string::npos) {
-                                FCITX_IBUS_DEBUG()
-                                    << "try to kill ibus-daemon.";
-                                // Well we can't kill it so better not to
-                                // replace it.
-                                if (kill(address.second, SIGKILL) != 0) {
-                                    return true;
-                                }
-                            }
-                        }
-                        becomeIBus();
-                        return true;
-                    });
-            }
-        }
-    }
-
-    if (!timeEvent_) {
-        becomeIBus();
+    if (address.first.empty() ||
+        address.first.find("fcitx_random_string") != std::string::npos) {
+        return std::nullopt;
     }
+    return address;
 }
 
-void IBusFrontendModule::becomeIBus() {
-    // ibusBus_.reset();
-    FCITX_IBUS_DEBUG() << "Requesting IBus service name.";
-    if (!bus()->requestName(
-            "org.freedesktop.IBus",
-            Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
-                                         dbus::RequestNameFlag::Queue})) {
-        FCITX_IBUS_DEBUG() << "Failed to request IBus service name.";
+std::optional<std::pair<std::string, pid_t>>
+readIBusInfo(const std::set<std::string> &socketPaths) {
+    for (const auto &path : socketPaths) {
+        auto address = readIBusInfo(path);
+        if (address) {
+            return address;
+        }
+    }
+    return std::nullopt;
+}
+
+void IBusFrontendModule::replaceIBus(bool recheck) {
+    if (retry_ <= 0) {
         return;
     }
+    // Ensure we don't dead loop here.
+    retry_ -= 1;
+    FCITX_IBUS_DEBUG() << "Found ibus socket files: " << socketPaths_;
 
-    inputMethod1_ = std::make_unique<IBusFrontend>(
-        this, bus(), IBUS_INPUTMETHOD_DBUS_INTERFACE);
+    if (isInFlatpak()) {
+        // Send exit request to all possible addresses.
+        bool deferCheck = false;
+        for (const auto &socketPath : socketPaths_) {
+            auto address = readIBusInfo(socketPath);
+            if (!address) {
+                continue;
+            }
+            const auto &oldAddress = address->first;
+            FCITX_IBUS_DEBUG() << "Old ibus address is: " << oldAddress;
+            FCITX_IBUS_DEBUG() << "Connecting to ibus address: " << oldAddress;
+            try {
+                dbus::Bus bus(oldAddress);
+                if (bus.isOpen()) {
+                    auto call = bus.createMethodCall(
+                        IBUS_SERVICE, "/org/freedesktop/IBus",
+                        IBUS_INPUTMETHOD_DBUS_INTERFACE, "Exit");
+                    call << false;
+                    call.call(1000000);
+                    deferCheck = true;
+                }
+            } catch (...) {
+            }
+        }
+        if (deferCheck) {
+            // Wait 1 second to become ibus.
+            timeEvent_ = instance()->eventLoop().addTimeEvent(
+                CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1000000, 0,
+                [this, recheck](EventSourceTime *, uint64_t) {
+                    becomeIBus(recheck);
+                    return true;
+                });
+            return;
+        }
+    } else if (auto optionalAddress = readIBusInfo(socketPaths_)) {
+        auto pid = runIBusExit();
+        if (pid > 0) {
+            FCITX_IBUS_DEBUG() << "Running ibus exit.";
+            timeEvent_ = instance()->eventLoop().addTimeEvent(
+                CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1000000, 0,
+                [this, pid, address = std::move(*optionalAddress),
+                 recheck](EventSourceTime *, uint64_t) {
+                    int stat = -1;
+                    pid_t ret;
+                    while ((ret = waitpid(pid, &stat, WNOHANG)) <= 0) {
+                        if (ret == 0) {
+                            FCITX_IBUS_DEBUG()
+                                << "ibus exit haven't ended yet, kill it.";
+                            kill(pid, SIGKILL);
+                            waitpid(pid, &stat, WNOHANG);
+                            break;
+                        }
+
+                        if (errno != EINTR) {
+                            stat = -1;
+                            break;
+                        }
+                    }
+
+                    FCITX_IBUS_DEBUG() << "ibus exit returns with " << stat;
+                    if (stat != 0) {
+                        // Re-read to ensure we have the latest information.
+                        auto newAddress = readIBusInfo(socketPaths_);
+                        if (address != newAddress) {
+                            // We should try ibus exit again.
+                            // This is not recursive because it's in time
+                            // event callback.
+                            replaceIBus(recheck);
+                            return true;
+                        }
+                        auto cmd = readFileContent(stringutils::joinPath(
+                            "/proc", address.second, "cmdline"));
+                        if (cmd.find("ibus-daemon") != std::string::npos) {
+                            FCITX_IBUS_DEBUG() << "try to kill ibus-daemon.";
+                            // Well we can't kill it so better not to
+                            // replace it.
+                            if (kill(address.second, SIGKILL) != 0) {
+                                return true;
+                            }
+                        }
+                    }
+                    becomeIBus(recheck);
+                    return true;
+                });
+            return;
+        }
+    }
+
+    becomeIBus(recheck);
+}
+
+void IBusFrontendModule::becomeIBus(bool recheck) {
+    // ibusBus_.reset();
     RawConfig config;
     auto address = bus()->address();
+    if (isInFlatpak()) {
+        if (address.find("/run/flatpak/bus") != std::string::npos) {
+            auto userBus = stringutils::joinPath(
+                standardPath_.userDirectory(StandardPath::Type::Runtime),
+                "bus");
+
+            struct stat statbuf;
+
+            if (::stat(userBus.data(), &statbuf) == 0 &&
+                (statbuf.st_mode & S_IFMT) == S_IFSOCK &&
+                statbuf.st_uid == getuid()) {
+                address = stringutils::concat("unix:path=", userBus);
+            }
+        }
+    }
     // This is a small hack to make ibus think that address is changed.
     // Otherwise it won't retry connection since we always use session bus
     // instead of start our own one.
@@ -821,7 +933,7 @@ void IBusFrontendModule::becomeIBus() {
     }
     address.append(",fcitx_random_string=");
     ICUUID uuid;
-    uuid_generate(uuid.data());
+    generateUUID(uuid.data());
     for (auto v : uuid) {
         address.append(fmt::format("{:02x}", static_cast<int>(v)));
     }
@@ -834,30 +946,61 @@ void IBusFrontendModule::becomeIBus() {
 
     FCITX_IBUS_DEBUG() << "Writing ibus daemon info.";
     for (const auto &path : socketPaths_) {
-        if (!StandardPath::global().safeSave(
+        if (!standardPath_.safeSave(
                 StandardPath::Type::Config, path,
                 [&config](int fd) { return writeAsIni(config, fd); })) {
             return;
         }
     }
 
-    addressWrote_ = address;
+    addressWrote_ = std::move(address);
     pidWrote_ = pidToWrite;
 
-    bus()->requestName(
-        IBUS_PANEL_SERVICE_NAME,
-        Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
-                                     dbus::RequestNameFlag::Queue});
+    if (!recheck) {
+        return;
+    }
 
-    portalBus_ = std::make_unique<dbus::Bus>(bus()->address());
-    portalIBusFrontend_ = std::make_unique<IBusFrontend>(
-        this, portalBus_.get(), IBUS_PORTAL_DBUS_INTERFACE);
-    portalBus_->attachEventLoop(&instance()->eventLoop());
-    if (!portalBus_->requestName(
-            IBUS_PORTAL_DBUS_SERVICE,
-            Flags<dbus::RequestNameFlag>{dbus::RequestNameFlag::ReplaceExisting,
-                                         dbus::RequestNameFlag::Queue})) {
-        FCITX_IBUS_WARN() << "Can not get portal ibus name right now.";
+    auto timeEvent = instance()->eventLoop().addTimeEvent(
+        CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 1500000, 0,
+        [this](EventSourceTime *, uint64_t) {
+            ensureIsIBus();
+            return true;
+        });
+
+    auto that = this;
+    that->timeEvent_ = std::move(timeEvent);
+}
+
+void IBusFrontendModule::ensureIsIBus() {
+    if (!isInFlatpak()) {
+        auto myname = bus()->uniqueName();
+        if (!myname.empty() &&
+            myname != bus()->serviceOwner(IBUS_SERVICE, 1000000)) {
+            auto msg = bus()->createMethodCall(
+                "org.freedesktop.DBus", "/org/freedesktop/DBus",
+                "org.freedesktop.DBus", "GetConnectionUnixProcessID");
+            msg << IBUS_SERVICE;
+            auto reply = msg.call(1000000);
+
+            uint32_t pid = 0;
+            if (reply.type() == dbus::MessageType::Reply) {
+                reply >> pid;
+            }
+            if (pid > 0 && static_cast<pid_t>(pid) != getpid()) {
+                if (kill(pid, SIGKILL) != 0) {
+                    return;
+                }
+            }
+        }
+    }
+
+    // Some one overwrite our address file.
+    for (const auto &path : socketPaths_) {
+        auto address = getAddress(path);
+        if (address.first != addressWrote_ || address.second != pidWrote_) {
+            replaceIBus(false);
+            return;
+        }
     }
 }
 
