@@ -5,15 +5,12 @@
  *
  */
 #include "quickphrase.h"
-#include <fcntl.h>
 
 #include <utility>
 #include "fcitx-config/iniparser.h"
-#include "fcitx-utils/charutils.h"
 #include "fcitx-utils/i18n.h"
 #include "fcitx-utils/inputbuffer.h"
-#include "fcitx-utils/standardpath.h"
-#include "fcitx-utils/utf8.h"
+#include "fcitx/addonfactory.h"
 #include "fcitx/addonmanager.h"
 #include "fcitx/candidatelist.h"
 #include "fcitx/inputcontextmanager.h"
@@ -105,7 +102,13 @@ QuickPhrase::QuickPhrase(Instance *instance)
 
             auto candidateList = inputContext->inputPanel().candidateList();
             if (candidateList) {
-                int idx = keyEvent.key().keyListIndex(selectionKeys_);
+                int idx = -1;
+                if (!selectionKeys_.empty() &&
+                    selectionKeys_[0].sym() == FcitxKey_1) {
+                    idx = keyEvent.key().digitSelection(selectionModifier_);
+                } else {
+                    idx = keyEvent.key().keyListIndex(selectionKeys_);
+                }
                 if (idx >= 0) {
                     if (idx < candidateList->size()) {
                         keyEvent.accept();
@@ -115,7 +118,7 @@ QuickPhrase::QuickPhrase(Instance *instance)
                 }
 
                 if (keyEvent.key().check(FcitxKey_space) &&
-                    candidateList->size()) {
+                    !candidateList->empty()) {
                     keyEvent.accept();
                     if (candidateList->cursorIndex() >= 0) {
                         candidateList->candidate(candidateList->cursorIndex())
@@ -143,14 +146,22 @@ QuickPhrase::QuickPhrase(Instance *instance)
 
                 if (keyEvent.key().checkKeyList(
                         instance_->globalConfig().defaultNextPage())) {
-                    keyEvent.filterAndAccept();
-                    candidateList->toPageable()->next();
-                    inputContext->updateUserInterface(
-                        UserInterfaceComponent::InputPanel);
-                    return;
+                    auto *pageable = candidateList->toPageable();
+                    if (!pageable->hasNext()) {
+                        if (pageable->usedNextBefore()) {
+                            event.accept();
+                            return;
+                        }
+                    } else {
+                        event.accept();
+                        pageable->next();
+                        inputContext->updateUserInterface(
+                            UserInterfaceComponent::InputPanel);
+                        return;
+                    }
                 }
 
-                if (candidateList->size() &&
+                if (!candidateList->empty() &&
                     keyEvent.key().checkKeyList(
                         instance_->globalConfig().defaultPrevCandidate())) {
                     keyEvent.filterAndAccept();
@@ -160,7 +171,7 @@ QuickPhrase::QuickPhrase(Instance *instance)
                     return;
                 }
 
-                if (candidateList->size() &&
+                if (!candidateList->empty() &&
                     keyEvent.key().checkKeyList(
                         instance_->globalConfig().defaultNextCandidate())) {
                     keyEvent.filterAndAccept();
@@ -210,6 +221,50 @@ QuickPhrase::QuickPhrase(Instance *instance)
                 keyEvent.accept();
                 return;
             }
+            if (keyEvent.key().check(FcitxKey_Delete)) {
+                if (state->buffer_.empty()) {
+                    state->reset(inputContext);
+                } else {
+                    if (state->buffer_.del()) {
+                        if (state->buffer_.empty()) {
+                            state->reset(inputContext);
+                        } else {
+                            updateUI(inputContext);
+                        }
+                    }
+                }
+                keyEvent.accept();
+                return;
+            }
+            if (!state->buffer_.empty()) {
+                const Key &key = keyEvent.key();
+                if (key.check(FcitxKey_Home) || key.check(FcitxKey_KP_Home)) {
+                    state->buffer_.setCursor(0);
+                    keyEvent.accept();
+                    return updateUI(inputContext);
+                }
+                if (key.check(FcitxKey_End) || key.check(FcitxKey_KP_End)) {
+                    state->buffer_.setCursor(state->buffer_.size());
+                    keyEvent.accept();
+                    return updateUI(inputContext);
+                }
+                if (key.check(FcitxKey_Left) || key.check(FcitxKey_KP_Left)) {
+                    auto cursor = state->buffer_.cursor();
+                    if (cursor > 0) {
+                        state->buffer_.setCursor(cursor - 1);
+                    }
+                    keyEvent.accept();
+                    return updateUI(inputContext);
+                }
+                if (key.check(FcitxKey_Right) || key.check(FcitxKey_KP_Right)) {
+                    auto cursor = state->buffer_.cursor();
+                    if (cursor < state->buffer_.size()) {
+                        state->buffer_.setCursor(cursor + 1);
+                    }
+                    keyEvent.accept();
+                    return updateUI(inputContext);
+                }
+            }
             if (!state->typed_ && !state->str_.empty() &&
                 state->buffer_.empty() && keyEvent.key().check(state->key_)) {
                 keyEvent.accept();
@@ -237,6 +292,30 @@ QuickPhrase::QuickPhrase(Instance *instance)
 
             updateUI(inputContext);
         }));
+    eventHandlers_.emplace_back(instance_->watchEvent(
+        EventType::InputContextInvokeAction, EventWatcherPhase::PreInputMethod,
+        [this](Event &event) {
+            auto &invokeActionEvent = static_cast<InvokeActionEvent &>(event);
+            auto *inputContext = invokeActionEvent.inputContext();
+            auto *state = inputContext->propertyFor(&factory_);
+            if (!state->enabled_) {
+                return;
+            }
+            invokeActionEvent.filter();
+            int cursor = invokeActionEvent.cursor() -
+                         static_cast<int>(state->prefix_.size());
+            if (cursor < 0 ||
+                invokeActionEvent.action() !=
+                    InvokeActionEvent::Action::LeftClick ||
+                !inputContext->capabilityFlags().test(
+                    CapabilityFlag::Preedit)) {
+                state->reset(inputContext);
+                return;
+            }
+            state->buffer_.setCursor(cursor);
+            invokeActionEvent.accept();
+            updateUI(inputContext);
+        }));
 
     reloadConfig();
 }
@@ -259,12 +338,8 @@ public:
 
             q_->updateUI(inputContext);
         } else if (action_ == QuickPhraseAction::Commit) {
-            state->reset(inputContext);
-            inputContext->inputPanel().reset();
-            inputContext->updatePreedit();
-            inputContext->updateUserInterface(
-                UserInterfaceComponent::InputPanel, true);
             inputContext->commitString(commit_);
+            state->reset(inputContext);
         }
         // DoNothing and other values are also handled here.
     }
@@ -299,24 +374,23 @@ void QuickPhrase::setSelectionKeys(QuickPhraseAction action) {
     }
 
     selectionKeys_.clear();
-
-    KeyStates states;
+    selectionModifier_ = KeyState::NoState;
     switch (config_.chooseModifier.value()) {
     case QuickPhraseChooseModifier::Alt:
-        states = KeyState::Alt;
+        selectionModifier_ = KeyState::Alt;
         break;
     case QuickPhraseChooseModifier::Control:
-        states = KeyState::Ctrl;
+        selectionModifier_ = KeyState::Ctrl;
         break;
     case QuickPhraseChooseModifier::Super:
-        states = KeyState::Super;
+        selectionModifier_ = KeyState::Super;
         break;
     default:
         break;
     }
 
     for (auto sym : syms) {
-        selectionKeys_.emplace_back(sym, states);
+        selectionKeys_.emplace_back(sym, selectionModifier_);
     }
 }
 
@@ -325,6 +399,8 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
     inputContext->inputPanel().reset();
     if (!state->buffer_.empty()) {
         auto candidateList = std::make_unique<CommonCandidateList>();
+        candidateList->setCursorPositionAfterPaging(
+            CursorPositionAfterPaging::ResetToFirst);
         candidateList->setPageSize(instance_->globalConfig().defaultPageSize());
         QuickPhraseProvider *providers[] = {&callbackProvider_,
                                             &builtinProvider_, &spellProvider_};
@@ -373,24 +449,23 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
 
         setSelectionKeys(selectionKeyAction);
         candidateList->setSelectionKey(selectionKeys_);
-        if (candidateList->size()) {
+        if (!candidateList->empty()) {
             candidateList->setGlobalCursorIndex(0);
         }
         inputContext->inputPanel().setCandidateList(std::move(candidateList));
     }
     Text preedit;
-    if (!state->prefix_.empty()) {
-        preedit.append(state->prefix_);
-    }
     const bool useClientPreedit =
         inputContext->capabilityFlags().test(CapabilityFlag::Preedit);
     TextFormatFlags format{useClientPreedit ? TextFormatFlag::Underline
                                             : TextFormatFlag::NoFlag};
-    preedit.append(state->buffer_.userInput(), format);
-    if (!state->buffer_.empty()) {
-        preedit.setCursor(state->prefix_.size() +
-                          state->buffer_.cursorByChar());
+    if (!state->prefix_.empty()) {
+        preedit.append(state->prefix_, format);
     }
+    if (!state->buffer_.empty()) {
+        preedit.append(state->buffer_.userInput(), format);
+    }
+    preedit.setCursor(state->prefix_.size() + state->buffer_.cursorByChar());
 
     Text auxUp(_("Quick Phrase: "));
     if (!state->typed_) {
@@ -398,7 +473,6 @@ void QuickPhrase::updateUI(InputContext *inputContext) {
     }
     inputContext->inputPanel().setAuxUp(auxUp);
     if (useClientPreedit) {
-        preedit.setCursor(0);
         inputContext->inputPanel().setClientPreedit(preedit);
     } else {
         inputContext->inputPanel().setPreedit(preedit);
