@@ -7,19 +7,9 @@
 #ifndef _FCITX_UI_CLASSIC_WAYLANDUI_H_
 #define _FCITX_UI_CLASSIC_WAYLANDUI_H_
 
-#include <cairo/cairo.h>
 #include "classicui.h"
-#include "config.h"
-#include "display.h"
+#include "waylandcursortheme.h"
 #include "waylandpointer.h"
-#include "wl_pointer.h"
-
-#ifdef CAIRO_EGL_FOUND
-
-#include <EGL/egl.h>
-#include <wayland-egl.h>
-
-#endif
 
 namespace fcitx {
 namespace classicui {
@@ -32,21 +22,7 @@ public:
     WaylandUI(ClassicUI *parent, const std::string &name, wl_display *display);
     ~WaylandUI();
 
-#ifdef CAIRO_EGL_FOUND
-
-    bool initEGL();
-    EGLSurface createEGLSurface(wl_egl_window *window,
-                                const EGLint *attrib_list);
-    void destroyEGLSurface(EGLSurface surface);
-    EGLContext argbCtx() { return argbCtx_; }
-    EGLDisplay eglDisplay() { return eglDisplay_; }
-
-    cairo_surface_t *createEGLCairoSurface(EGLSurface surface, int width,
-                                           int height);
-#endif
-
     ClassicUI *parent() const { return parent_; }
-    const std::string &name() const { return name_; }
     wayland::Display *display() const { return display_; }
     void update(UserInterfaceComponent component,
                 InputContext *inputContext) override;
@@ -56,25 +32,18 @@ public:
 
     std::unique_ptr<WaylandWindow> newWindow();
 
+    WaylandCursorTheme *cursorTheme() { return cursorTheme_.get(); }
+
 private:
     void setupInputWindow();
 
     ClassicUI *parent_;
-    std::string name_;
     wayland::Display *display_;
     ScopedConnection panelConn_, panelRemovedConn_;
+    std::unique_ptr<WaylandCursorTheme> cursorTheme_;
     std::unique_ptr<WaylandPointer> pointer_;
     std::unique_ptr<WaylandInputWindow> inputWindow_;
-
-    bool isSuspend_ = true;
-    bool hasEgl_ = false;
-#ifdef CAIRO_EGL_FOUND
-    // EGL stuff
-    EGLDisplay eglDisplay_ = nullptr;
-    EGLConfig argbConfig_ = nullptr;
-    EGLContext argbCtx_ = nullptr;
-    cairo_device_t *argbDevice_ = nullptr;
-#endif
+    std::unique_ptr<EventSource> defer_;
 };
 } // namespace classicui
 } // namespace fcitx
